@@ -1,12 +1,14 @@
-## plotAndPlayGTK: interactive plots in R using GTK
+## plotAndPlayGTK: interactive plots in R using GTK+
 ##
-## Copyright (c) 2007 Felix Andrews <felix@nfrac.org>, GPL
+## Copyright (c) 2007 Felix Andrews <felix@nfrac.org> and others
+## GPL version 2 or newer
 
 MAJOR <- "0"
 MINOR <- "7"
-REVISION <- unlist(strsplit("$Revision: 17 $", split=" "))[2]
+REVISION <- unlist(strsplit("$Revision: 19 $", split=" "))[2]
 VERSION <- paste(MAJOR, MINOR, REVISION, sep=".")
-COPYRIGHT <- "(c) 2007 Felix Andrews <felix@nfrac.org>, GPL"
+COPYRIGHT <- paste("(c) 2007 Felix Andrews <felix@nfrac.org>",
+	"with contributions from Graham Williams and Deepayan Sarkar")
 WEBSITE <- "http://code.google.com/p/plotandplay-gtk/"
 
 ## LICENSE
@@ -25,39 +27,35 @@ WEBSITE <- "http://code.google.com/p/plotandplay-gtk/"
 ## along with this program; if not, write to the Free Software
 ## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-.StateEnv <- new.env()
-.StateEnv$win <- list()
-.StateEnv$dev <- list()
-.StateEnv$call <- list()
-.StateEnv$id.call <- list()
-.StateEnv$focus <- list()
-.StateEnv$nav.scales <- list()
-.StateEnv$trans.scales <- list()
-.StateEnv$is.lattice <- list()
-.StateEnv$page <- list()
+StateEnv <- new.env()
 
 plotAndPlayButtons <- alist(
 	identify=list("Identify data", "gtk-info", f=.plotAndPlay_identify_event),
-	zoomin=list("Zoom in", "gtk-zoom-in", f=.plotAndPlay_zoomin_event),
+	zoomin=list("Zoom", "gtk-zoom-in", f=.plotAndPlay_zoomin_event),
 	zoomout=list("Zoom out", "gtk-zoom-out", f=.plotAndPlay_zoomout_event),
 	zoomfit=list("Zoom to fit", "gtk-zoom-fit", f=.plotAndPlay_zoomfit_event),
 	centre=list("Re-centre", "gtk-jump-to-ltr", f=.plotAndPlay_centre_event),
 	zero=list("Full scale", "gtk-goto-bottom", f=.plotAndPlay_zero_event, isToggle=T),
 	logscale=list("Log scale", "gtk-goto-top", f=.plotAndPlay_logscale_event, isToggle=T),
 	focus=list("Choose panel", "gtk-select-color", f=.plotAndPlay_focus_event),
+	expand=list("Expand panel", "gtk-fullscreen", f=.plotAndPlay_expand_event, isToggle=T),
 	prev.page=list("Prev page", "gtk-go-back-ltr", f=.plotAndPlay_prevpage_event),
-	next.page=list("Next page", "gtk-go-forward-ltr", f=.plotAndPlay_nextpage_event)
+	next.page=list("Next page", "gtk-go-forward-ltr", f=.plotAndPlay_nextpage_event),
+	brush=list("Brush", "gtk-media-record", f=.plotAndPlay_brush_event)
 )
 
 #gdkPixbufGetFromDrawable(dest = NULL, src, cmap = NULL, src.x, src.y, dest.x, dest.y, width, height)
 #"gtk-index" / "gtk-preferences"
-#"gtk-zoom-100" (relation="iso")
+#"gtk-find"
+#"gtk-italic" (edit labels/titles)
+#"gtk-undo-ltr"
 
 latticeNames <- c("barchart", "bwplot", "cloud", "contourplot", "densityplot", 
 	"dotplot", "histogram", "levelplot", "parallel", "qq", "qqmath", "rfs", 
 	"splom", "stripplot", "tmd", "wireframe", "xyplot")
 
-plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), trans.scales=c("y"), buttons=plotAndPlayButtons[c("identify", "zoomin", "zoomout", "zoomfit", "centre")], extra.buttons=plotAndPlayButtons[c("zero")], labels=NULL, identify.call=NULL, is.lattice=(callName %in% latticeNames), eval.args.pattern=".", envir=parent.frame()) {
+
+plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), trans.scales=c("y"), buttons=plotAndPlayButtons[c("identify", "zoomin", "zoomout", "zoomfit", "centre")], extra.buttons=plotAndPlayButtons[c("zero")], labels=NULL, identify.call=NULL, is.lattice=(callName %in% latticeNames), eval.args=F, envir=parent.frame()) {
 	if (!require("cairoDevice", quietly=TRUE)) {
 		stop("Require package 'cairoDevice'")
 	}
@@ -77,9 +75,13 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	}
 	# put call into canonical form and evaluate its arguments
 	plot.call <- match.call(eval(plot.call[[1]]), plot.call)
-	plot.call <- evalCallArgs(plot.call, envir=envir, pattern=eval.args.pattern)
+	if (!identical(eval.args, FALSE)) {
+		plot.call <- evalCallArgs(plot.call, envir=envir, pattern=eval.args)
+	}
 	callName <- deparse(plot.call[[1]])
-	if (is.lattice) {
+	if (is.lattice &&
+		!(callName %in% c("splom", "cloud", "levelplot",
+			"contourplot", "wireframe", "parallel")) ) {
 		# need this for correctly identifying points
 		plot.call$subscripts <- TRUE
 	}
@@ -91,8 +93,9 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 			# try to guess labels
 			if (missing(labels)) {
 				if ('data' %in% names(plot.call)) {
-					labels <- row.names(plot.call$data)
+					labels <- row.names(eval(plot.call$data))
 				}
+				# TODO: try to get row names from formula
 			}
 			if (!is.null(labels)) {
 				identify.call$labels <- bquote(
@@ -112,10 +115,11 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 			# try to guess labels
 			if (missing(labels)) {
 				if ('x' %in% names(plot.call)) {
-					labels <- row.names(plot.call$x)
+					tmp.x <- eval(plot.call$x)
+					labels <- row.names(tmp.x)
 					if (is.null(labels)) {
-						if (is.ts(plot.call$x)) {
-							labels <- time(x)
+						if (is.ts(tmp.x)) {
+							labels <- time(tmp.x)
 						}
 					}
 				}
@@ -124,10 +128,10 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 				identify.call$labels <- labels
 			}
 		}
-		identify.call$cex <- 0.75
+		identify.call$cex <- 0.7
 	}
 	# check whether the window already exists
-	if (is.null(.StateEnv$win[[name]])) {
+	if (is.null(StateEnv[[name]])) {
 		# create a new GTK window and set up cairoDevice for plotting
 		myWin <- gtkWindow(show=FALSE)
 		myWin[["default-width"]] <- 600
@@ -144,22 +148,39 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 		trellis.device(new=F)
 	} else {
 		# window exists
-		myWin <- .StateEnv$win[[name]]
+		myWin <- StateEnv[[name]]$win
 		myVBox <- myWin$getChildren()[[1]]
 		# switch to the device
-		dev.set(.StateEnv$dev[[name]])
+		dev.set(StateEnv[[name]]$dev)
 		# blank the plot
-		.StateEnv$call[[name]] <- NULL
+		StateEnv[[name]]$call <- NULL
 		plot.new()
 		# remove the toolbar, it will be recreated
 		plotAndPlayGetToolbar(name)$destroy()
 	}
 	# set which buttons are visible
+	if (is.lattice &&
+		(callName %in% c("cloud", "wireframe")) ) {
+		if (missing(buttons)) {
+			buttons <- list()
+		}
+	}
+	if (is.lattice &&
+		(callName %in% c("splom")) ) {
+		if (missing(buttons)) {
+			buttons <- plotAndPlayButtons[c("brush")]
+		}
+		if (missing(extra.buttons)) {
+			extra.buttons <- list()
+		}
+	}
+	extra.unique <- !(names(extra.buttons) %in% names(buttons))
+	buttons <- c(buttons, extra.buttons[extra.unique])
 	if (is.lattice) {
 		result <- eval(plot.call)
 		nPanels <- prod(dim(result))
 		nPerPage <- nPanels
-		myLayout <- plot.call$layout
+		myLayout <- eval(plot.call$layout)
 		if (!is.null(myLayout)) {
 			nPerPage <- myLayout[1] * myLayout[2]
 			if (myLayout[1] == 0) { nPerPage <- myLayout[2] }
@@ -168,13 +189,13 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 				buttons$next.page <- plotAndPlayButtons$next.page
 			}
 			# needed for page navigation (packet.panel.page)
-			plot.call$layout[3] <- 1
+			myLayout[3] <- 1
+			plot.call$layout <- myLayout
 		}
 		if (nPerPage > 1) {
-			buttons$focus <- plotAndPlayButtons$focus
+			buttons$expand <- plotAndPlayButtons$expand
 		}
 	}
-	buttons <- c(buttons, extra.buttons)
 	theToolbar <- gtkToolbar()
 	theToolbar[["toolbar-style"]] <- GtkToolbarStyle['both']
 	myVBox$packStart(theToolbar, expand=FALSE)
@@ -227,31 +248,30 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	# bring window to front
 	myWin$present()
 	# store the state of this plot window
-	.StateEnv$win[[name]] <- myWin
-	.StateEnv$dev[[name]] <- dev.cur()
-	.StateEnv$call[[name]] <- plot.call
-	.StateEnv$id.call[[name]] <- identify.call
-	.StateEnv$focus[[name]] <- list(col=0, row=0)
-	.StateEnv$nav.scales[[name]] <- nav.scales
-	.StateEnv$trans.scales[[name]] <- trans.scales
-	.StateEnv$is.lattice[[name]] <- is.lattice
-	.StateEnv$page[[name]] <- 1
+	StateEnv[[name]] <- list(
+		win=myWin,
+		dev=dev.cur(),
+		call=plot.call,
+		id.call=identify.call,
+		focus=list(col=0, row=0),
+		nav.scales=nav.scales,
+		trans.scales=trans.scales,
+		is.lattice=is.lattice,
+		page=1
+	)
 	# do the plot
-	result <- eval(plot.call)
-	if (identical(class(result), "trellis")) { print(result) }
-	invisible(result)
+	invisible(plotAndPlayUpdate(name))
 }
 
 .plotAndPlay_close_event <- function(widget, event, user.data) {
 	name <- user.data$name
-	if (!is.null(.StateEnv$win[[name]])) {
-		.StateEnv$win[[name]]$destroy()
+	if (StateEnv[[name]]$dev %in% dev.list()) {
+		dev.off(StateEnv[[name]]$dev)
 	}
-	.StateEnv$win[[name]] <- NULL
-	.StateEnv$dev[[name]] <- NULL
-	.StateEnv$call[[name]] <- NULL
-	.StateEnv$id.call[[name]] <- NULL
-	.StateEnv$focus[[name]] <- NULL
+	if (!is.null(StateEnv[[name]]$win)) {
+		StateEnv[[name]]$win$destroy()
+	}
+	StateEnv[[name]] <- NULL
 }
 
 .plotAndPlay_save_event <- function(widget, user.data) {
@@ -261,7 +281,7 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# get filename
 	myExt <- if (!is.null(user.data$ext)) { user.data$ext } else { 'pdf' }
@@ -270,7 +290,7 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 		filters=Filters[c("pdf","png","ps","All"),],
 		index=match(myExt, c("pdf","png","ps","svg"))
 	)
-	.StateEnv$win[[name]]$present()
+	StateEnv[[name]]$win$present()
 	if (is.na(filename)) { return() }
 	ext <- tolower(get.extension(filename))
 	if (ext == "") {
@@ -313,7 +333,7 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# save plot to file
 	filename <- paste(tempfile(), ".png", sep="")
@@ -337,7 +357,7 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# get new greyscale setting
 	greyscale <- widget$getActive()
@@ -352,14 +372,14 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 
 .plotAndPlay_zoomin_event <- function(widget, user.data) {
 	name <- user.data$name
-	nav.x <- ("x" %in% .StateEnv$nav.scales[[name]])
-	nav.y <- ("y" %in% .StateEnv$nav.scales[[name]])
+	nav.x <- ("x" %in% StateEnv[[name]]$nav.scales)
+	nav.y <- ("y" %in% StateEnv[[name]]$nav.scales)
 	# disable other plot buttons until this is over
 	plotAndPlayGetToolbar(name)$setSensitive(F)
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# set up prompt
 	plotAndPlayMakePrompt(name)
@@ -370,21 +390,21 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	highEdge <- "top-right corner"
 	if (!nav.y) { highEdge <- "right edge" }
 	if (!nav.x) { highEdge <- "top edge" }
-	plotAndPlaySetPrompt(name, paste("Click at the", lowEdge, 
-		"of the region to zoom in to"))
 	# set up masking
 	maskGrob <- rectGrob(gp=gpar(col="grey", 
 		fill=rgb(0.5,0.5,0.5, alpha=0.5)), name="tmp.mask")
 	# get new scales interactively
-	if (.StateEnv$is.lattice[[name]]) {
+	if (StateEnv[[name]]$is.lattice) {
 		# lattice plot
-		if (!any(.StateEnv$focus[[name]])) { trellis.focus("panel", 1, 1) }
-		xlim <- as.numeric(convertX(unit(0:1, "npc"), "native"))
-		ylim <- as.numeric(convertY(unit(0:1, "npc"), "native"))
+		plotAndPlayDoFocus(name)
+		plotAndPlaySetPrompt(name, paste("Click at the", lowEdge, 
+			"of the region to zoom in to"))
+		xlim <- convertX(unit(0:1, "npc"), "native", valueOnly=T)
+		ylim <- convertY(unit(0:1, "npc"), "native", valueOnly=T)
 		# get lower limits
 		clickLoc <- grid.locator()
 		if (is.null(clickLoc)) {
-			if (!any(.StateEnv$focus[[name]])) { trellis.unfocus() }
+			if (!any(StateEnv[[name]]$focus)) { trellis.unfocus() }
 			return()
 		}
 		clickLoc <- lapply(clickLoc, as.numeric)
@@ -422,11 +442,13 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 			just=c("right", "top")))
 		# convert back from log scale if required
 		newlims <- latticeUnLog(xlim.new, ylim.new, 
-			.StateEnv$call[[name]]$scales)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- newlims$x }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- newlims$y }
+			StateEnv[[name]]$call$scales)
+		if (nav.x) { StateEnv[[name]]$call$xlim <- newlims$x }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- newlims$y }
 	} else {
 		# traditional graphics plot
+		plotAndPlaySetPrompt(name, paste("Click at the", lowEdge, 
+			"of the region to zoom in to"))
 		xlim <- par("usr")[1:2]
 		ylim <- par("usr")[3:4]
 		# get lower limits
@@ -449,38 +471,38 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 		if (nav.x) { abline(v=xlim.new[2], col="red") }
 		if (nav.y) { abline(h=ylim.new[2], col="red") }
 		# convert back from log scale if required (TODO)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- xlim.new }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- ylim.new }
+		if (nav.x) { StateEnv[[name]]$call$xlim <- xlim.new }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- ylim.new }
 	}
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_zoomout_event <- function(widget, user.data) {
 	name <- user.data$name
-	nav.x <- ("x" %in% .StateEnv$nav.scales[[name]])
-	nav.y <- ("y" %in% .StateEnv$nav.scales[[name]])
+	nav.x <- ("x" %in% StateEnv[[name]]$nav.scales)
+	nav.y <- ("y" %in% StateEnv[[name]]$nav.scales)
 	# disable other plot buttons until this is over
 	plotAndPlayGetToolbar(name)$setSensitive(F)
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# find existing scales and update call
-	if (.StateEnv$is.lattice[[name]]) {
+	if (StateEnv[[name]]$is.lattice) {
 		# lattice plot
-		if (!any(.StateEnv$focus[[name]])) {
+		if (!any(StateEnv[[name]]$focus)) {
 			trellis.focus("panel", 1, 1, highlight=F)
 		}
-		xlim <- as.numeric(convertX(unit(0:1, "npc"), "native"))
-		ylim <- as.numeric(convertY(unit(0:1, "npc"), "native"))
+		xlim <- convertX(unit(0:1, "npc"), "native", valueOnly=T)
+		ylim <- convertY(unit(0:1, "npc"), "native", valueOnly=T)
 		# zoom out: make range twice the size
 		if (nav.x) { xlim <- xlim + diff(xlim) * c(-0.5, 0.5) }
 		if (nav.y) { ylim <- ylim + diff(ylim) * c(-0.5, 0.5) }
 		# convert back from log scale if required
-		newlims <- latticeUnLog(xlim, ylim, .StateEnv$call[[name]]$scales)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- newlims$x }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- newlims$y }
+		newlims <- latticeUnLog(xlim, ylim, StateEnv[[name]]$call$scales)
+		if (nav.x) { StateEnv[[name]]$call$xlim <- newlims$x }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- newlims$y }
 	} else {
 		# traditional graphics plot
 		xlim <- par("usr")[1:2]
@@ -489,68 +511,69 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 		if (nav.x) { xlim <- xlim + diff(xlim) * c(-0.5, 0.5) }
 		if (nav.y) { ylim <- ylim + diff(ylim) * c(-0.5, 0.5) }
 		# convert back from log scale if required (TODO)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- xlim }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- ylim }
+		if (nav.x) { StateEnv[[name]]$call$xlim <- xlim }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- ylim }
 	}
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_zoomfit_event <- function(widget, user.data) {
 	name <- user.data$name
-	nav.x <- ("x" %in% .StateEnv$nav.scales[[name]])
-	nav.y <- ("y" %in% .StateEnv$nav.scales[[name]])
+	nav.x <- ("x" %in% StateEnv[[name]]$nav.scales)
+	nav.y <- ("y" %in% StateEnv[[name]]$nav.scales)
 	# disable other plot buttons until this is over
 	plotAndPlayGetToolbar(name)$setSensitive(F)
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# find existing scales and update call
-	if (.StateEnv$is.lattice[[name]]) {
+	if (StateEnv[[name]]$is.lattice) {
 		# lattice plot
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- NULL }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- NULL }
+		if (nav.x) { StateEnv[[name]]$call$xlim <- NULL }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- NULL }
 	} else {
 		# traditional graphics plot
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- NULL }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- NULL }
+		if (nav.x) { StateEnv[[name]]$call$xlim <- NULL }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- NULL }
 	}
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_centre_event <- function(widget, user.data) {
 	name <- user.data$name
-	nav.x <- ("x" %in% .StateEnv$nav.scales[[name]])
-	nav.y <- ("y" %in% .StateEnv$nav.scales[[name]])
+	nav.x <- ("x" %in% StateEnv[[name]]$nav.scales)
+	nav.y <- ("y" %in% StateEnv[[name]]$nav.scales)
 	# disable other plot buttons until this is over
 	plotAndPlayGetToolbar(name)$setSensitive(F)
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# set up prompt
 	plotAndPlayMakePrompt(name)
 	on.exit(plotAndPlayUnmakePrompt(name), add=T)
-	plotAndPlaySetPrompt(name, "Click to re-centre the plot")
 	# get new scales interactively
-	if (.StateEnv$is.lattice[[name]]) {
+	if (StateEnv[[name]]$is.lattice) {
 		# lattice plot
-		if (!any(.StateEnv$focus[[name]])) { trellis.focus("panel", 1, 1) }
-		xlim <- as.numeric(convertX(unit(0:1, "npc"), "native"))
-		ylim <- as.numeric(convertY(unit(0:1, "npc"), "native"))
+		plotAndPlayDoFocus(name)
+		plotAndPlaySetPrompt(name, "Click to re-centre the plot")
+		xlim <- convertX(unit(0:1, "npc"), "native", valueOnly=T)
+		ylim <- convertY(unit(0:1, "npc"), "native", valueOnly=T)
 		# get new centre point
 		clickLoc <- grid.locator()
 		if (is.null(clickLoc)) {
-			if (!any(.StateEnv$focus[[name]])) { trellis.unfocus() }
+			if (!any(StateEnv[[name]]$focus)) { trellis.unfocus() }
 			return()
 		}
 		clickLoc <- lapply(clickLoc, as.numeric)
 		if (nav.x) { xlim <- clickLoc$x + diff(xlim) * c(-0.5, 0.5) }
 		if (nav.y) { ylim <- clickLoc$y + diff(ylim) * c(-0.5, 0.5) }
 		# convert back from log scale if required
-		newlims <- latticeUnLog(xlim, ylim, .StateEnv$call[[name]]$scales)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- newlims$x }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- newlims$y }
+		newlims <- latticeUnLog(xlim, ylim, StateEnv[[name]]$call$scales)
+		if (nav.x) { StateEnv[[name]]$call$xlim <- newlims$x }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- newlims$y }
 	} else {
 		# traditional graphics plot
+		plotAndPlaySetPrompt(name, "Click to re-centre the plot")
 		xlim <- par("usr")[1:2]
 		ylim <- par("usr")[3:4]
 		# get new centre point
@@ -561,8 +584,8 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 		if (nav.x) { xlim <- clickLoc$x + diff(xlim) * c(-0.5, 0.5) }
 		if (nav.y) { ylim <- clickLoc$y + diff(ylim) * c(-0.5, 0.5) }
 		# convert back from log scale if required (TODO)
-		if (nav.x) { .StateEnv$call[[name]]$xlim <- xlim }
-		if (nav.y) { .StateEnv$call[[name]]$ylim <- ylim }
+		if (nav.x) { StateEnv[[name]]$call$xlim <- xlim }
+		if (nav.y) { StateEnv[[name]]$call$ylim <- ylim }
 	}
 	plotAndPlayUpdate(name)
 }
@@ -574,19 +597,51 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# set up prompt
 	plotAndPlayMakePrompt(name)
 	on.exit(plotAndPlayUnmakePrompt(name), add=T)
 	plotAndPlaySetPrompt(name, "Click on a panel to focus on (for further interaction)")
 	newFocus <- trellis.clickFocus()
-	plotAndPlaySetPrompt(name, "")
 	if (!is.null(newFocus)) {
-		.StateEnv$focus[[name]] <- newFocus
+		StateEnv[[name]]$focus <- newFocus
 	}
 }
 
+.plotAndPlay_expand_event <- function(widget, user.data) {
+	name <- user.data$name
+	# disable other plot buttons until this is over
+	plotAndPlayGetToolbar(name)$setSensitive(F)
+	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
+	# switch to this device
+	oldDev <- dev.cur()
+	dev.set(StateEnv[[name]]$dev)
+	on.exit(dev.set(oldDev), add=T)
+	# get new expanded setting
+	doExpand <- widget$getActive()
+	if (doExpand) {
+		# set up prompt
+		plotAndPlayMakePrompt(name)
+		on.exit(plotAndPlayUnmakePrompt(name), add=T)
+		plotAndPlaySetPrompt(name, "Click on a panel to expand (for further interaction)")
+		newFocus <- trellis.clickFocus(highlight=F)
+		if (is.null(newFocus)) {
+			return()
+		}
+		StateEnv[[name]]$focus <- list(col=0, row=0)
+		StateEnv[[name]]$old.call.layout <- StateEnv[[name]]$call$layout
+		StateEnv[[name]]$call$layout <- c(0,1,1)
+		StateEnv[[name]]$old.page <- StateEnv[[name]]$page
+		StateEnv[[name]]$page <- packet.number()
+	} else {
+		StateEnv[[name]]$focus <- list(col=0, row=0)
+		StateEnv[[name]]$call$layout <- StateEnv[[name]]$old.call.layout
+		StateEnv[[name]]$page <- StateEnv[[name]]$old.page
+	}
+	plotAndPlayUpdate(name)
+}
+
 .plotAndPlay_identify_event <- function(widget, user.data) {
 	name <- user.data$name
 	# disable other plot buttons until this is over
@@ -594,40 +649,57 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# set up prompt
 	plotAndPlayMakePrompt(name)
 	on.exit(plotAndPlayUnmakePrompt(name), add=T)
-	plotAndPlaySetPrompt(name, paste("Identifying data points...",
-		"Click the right mouse button to finish."))
 	# do identify
-	idCall <- .StateEnv$id.call[[name]]
-	if (.StateEnv$is.lattice[[name]]) {
+	idCall <- StateEnv[[name]]$id.call
+	if (StateEnv[[name]]$is.lattice) {
 		# lattice plot
-		if (!any(.StateEnv$focus[[name]])) {
-			if (length(trellis.currentLayout()) == 1) {
-				trellis.focus("panel", 1, 1)
-			} else {
-				plotAndPlaySetPrompt(name, "First, choose a panel")
-				trellis.clickFocus()
-			}
-			plotAndPlaySetPrompt(name, paste("Identifying data points...",
-				"Click the right mouse button to finish."))
-		}
+		plotAndPlayDoFocus(name, clip.off=T) #, highlight=F
+		plotAndPlaySetPrompt(name, paste("Identifying data points...",
+			"Click the right mouse button to finish."))
 		ids <- eval(idCall)
 		# TODO: store this to maintain state?
-		if (!any(.StateEnv$focus[[name]])) { trellis.unfocus() }
+		if (!any(StateEnv[[name]]$focus)) { trellis.unfocus() }
 	} else {
 		# traditional graphics plot
+		plotAndPlaySetPrompt(name, paste("Identifying data points...",
+			"Click the right mouse button to finish."))
 		ids <- eval(idCall)
 	}
 }
 
+.plotAndPlay_brush_event <- function(widget, user.data) {
+	name <- user.data$name
+	if (!StateEnv[[name]]$is.lattice) {
+		errorDialog("Brushing only works for Lattice plots (splom).")
+		return()
+	}
+	# disable other plot buttons until this is over
+	plotAndPlayGetToolbar(name)$setSensitive(F)
+	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
+	# switch to this device
+	oldDev <- dev.cur()
+	dev.set(StateEnv[[name]]$dev)
+	on.exit(dev.set(oldDev), add=T)
+	# set up prompt
+	plotAndPlayMakePrompt(name)
+	on.exit(plotAndPlayUnmakePrompt(name), add=T)
+	# do brushing
+	plotAndPlayDoFocus(name) #, highlight=F
+	plotAndPlaySetPrompt(name, paste("Brushing data points...",
+		"Click the right mouse button to finish."))
+	while(highlightOne()) {} # right click to exit
+	if (!any(StateEnv[[name]]$focus)) { trellis.unfocus() }
+}
+
 .plotAndPlay_zero_event <- function(widget, user.data) {
 	name <- user.data$name
-	trans.x <- ("x" %in% .StateEnv$trans.scales[[name]])
-	trans.y <- ("y" %in% .StateEnv$trans.scales[[name]])
+	trans.x <- ("x" %in% StateEnv[[name]]$trans.scales)
+	trans.y <- ("y" %in% StateEnv[[name]]$trans.scales)
 	isX <- isTRUE(user.data$x)
 	isY <- isTRUE(user.data$y)
 	if (!isX && !isY) { isX <- isY <- T }
@@ -636,19 +708,19 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# switch to this device
 	oldDev <- dev.cur()
-	dev.set(.StateEnv$dev[[name]])
+	dev.set(StateEnv[[name]]$dev)
 	on.exit(dev.set(oldDev), add=T)
 	# get new zero scale setting
 	zeroScale <- widget$getActive()
 	# make change and re-draw plot
 	if (zeroScale) {
-		if (.StateEnv$is.lattice[[name]]) {
+		if (StateEnv[[name]]$is.lattice) {
 			# lattice plot
-			if (!any(.StateEnv$focus[[name]])) {
+			if (!any(StateEnv[[name]]$focus)) {
 				trellis.focus("panel", 1, 1, highlight=F)
 			}
-			xlim <- as.numeric(convertX(unit(0:1, "npc"), "native"))
-			ylim <- as.numeric(convertY(unit(0:1, "npc"), "native"))
+			xlim <- convertX(unit(0:1, "npc"), "native", valueOnly=T)
+			ylim <- convertY(unit(0:1, "npc"), "native", valueOnly=T)
 		} else {
 			# traditional graphics plot
 			xlim <- par("usr")[1:2]
@@ -660,7 +732,7 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 			} else if (max(xlim) < 0) {
 				xlim[which.max(xlim)] <- 0
 			}
-			.StateEnv$call[[name]]$xlim <- xlim
+			StateEnv[[name]]$call$xlim <- xlim
 		}
 		if (trans.y) {
 			if (min(ylim) > 0) {
@@ -668,73 +740,85 @@ plotAndPlay <- function(expr, name="plot", plot.call, nav.scales=c("x","y"), tra
 			} else if (max(ylim) < 0) {
 				ylim[which.max(ylim)] <- 0
 			}
-			.StateEnv$call[[name]]$ylim <- ylim
+			StateEnv[[name]]$call$ylim <- ylim
 		}
 	} else {
-		if (trans.x) { .StateEnv$call[[name]]$xlim <- NULL }
-		if (trans.y) { .StateEnv$call[[name]]$ylim <- NULL }
+		if (trans.x) { StateEnv[[name]]$call$xlim <- NULL }
+		if (trans.y) { StateEnv[[name]]$call$ylim <- NULL }
 	}
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_logscale_event <- function(widget, user.data) {
 	name <- user.data$name
-	trans.x <- ("x" %in% .StateEnv$trans.scales[[name]])
-	trans.y <- ("y" %in% .StateEnv$trans.scales[[name]])
+	trans.x <- ("x" %in% StateEnv[[name]]$trans.scales)
+	trans.y <- ("y" %in% StateEnv[[name]]$trans.scales)
 	# disable other plot buttons until this is over
 	plotAndPlayGetToolbar(name)$setSensitive(F)
 	on.exit(plotAndPlayGetToolbar(name)$setSensitive(T))
 	# get new log scale setting
 	logScale <- widget$getActive()
 	# make change and re-draw plot
-	if (.StateEnv$is.lattice[[name]]) {
+	if (StateEnv[[name]]$is.lattice) {
 		if (trans.x && trans.y) {
 			# apply to both scales
-			.StateEnv$call[[name]]$scales$log <- logScale
+			StateEnv[[name]]$call$scales$log <- logScale
 		} else {
-			if (trans.x) { .StateEnv$call[[name]]$scales$x$log <- logScale }
-			if (trans.y) { .StateEnv$call[[name]]$scales$y$log <- logScale }
+			if (trans.x) { StateEnv[[name]]$call$scales$x$log <- logScale }
+			if (trans.y) { StateEnv[[name]]$call$scales$y$log <- logScale }
 		}
 	} else {
 		logSpec <- "xy"
 		if (!trans.y) { logSpec <- "x" }
 		if (!trans.x) { logSpec <- "y" }
-		.StateEnv$call[[name]]$log <- if (logScale) { logSpec }
+		StateEnv[[name]]$call$log <- if (logScale) { logSpec }
 	}
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_prevpage_event <- function(widget, user.data) {
 	name <- user.data$name
-	if (.StateEnv$page[[name]] == 1) { return() }
-	.StateEnv$page[[name]] <- .StateEnv$page[[name]] - 1
+	if (StateEnv[[name]]$page == 1) { return() }
+	StateEnv[[name]]$page <- StateEnv[[name]]$page - 1
 	plotAndPlayUpdate(name)
 }
 
 .plotAndPlay_nextpage_event <- function(widget, user.data) {
 	name <- user.data$name
-	.StateEnv$page[[name]] <- .StateEnv$page[[name]] + 1
+	StateEnv[[name]]$page <- StateEnv[[name]]$page + 1
 	plotAndPlayUpdate(name)
 }
 
-plotAndPlayUpdate <- function(name) {
-	result <- eval(.StateEnv$call[[name]])
-	if ("trellis" %in% class(result)) {
-		curPage <- .StateEnv$page[[name]]
-		print(result, packet.panel=packet.panel.page(curPage))
-		if (any(.StateEnv$focus[[name]])) {
-			with(.StateEnv$focus[[name]], 
-				trellis.focus("panel", col, row))
+plotAndPlayDoFocus <- function(name, highlight=T, ...) {
+	if (!any(StateEnv[[name]]$focus)) {
+		if (length(trellis.currentLayout()) == 1) {
+			trellis.focus("panel", 1, 1, highlight=F, ...)
+		} else {
+			plotAndPlaySetPrompt(name, "First, choose a panel")
+			trellis.clickFocus(highlight=highlight, ...)
 		}
 	}
 }
 
+plotAndPlayUpdate <- function(name) {
+	result <- eval(StateEnv[[name]]$call)
+	if (inherits(result, "trellis")) {
+		curPage <- StateEnv[[name]]$page
+		print(result, packet.panel=packet.panel.page(curPage))
+		if (any(StateEnv[[name]]$focus)) {
+			with(StateEnv[[name]]$focus, 
+				trellis.focus("panel", col, row))
+		}
+	}
+	invisible(result)
+}
+
 plotAndPlayGetDA <- function(name) {
-	.StateEnv$win[[name]]$getChildren()[[1]]$getChildren()[[1]]
+	StateEnv[[name]]$win$getChildren()[[1]]$getChildren()[[1]]
 }
 
 plotAndPlayGetToolbar <- function(name) {
-	.StateEnv$win[[name]]$getChildren()[[1]]$getChildren()[[2]]
+	StateEnv[[name]]$win$getChildren()[[1]]$getChildren()[[2]]
 }
 
 plotAndPlayMakePrompt <- function(name) {
@@ -771,7 +855,7 @@ plotAndPlaySetPrompt <- function(name, text) {
 	myLabel$setMarkup(paste(sep='','<big><b>', text, '</b></big>'))
 }
 
-trellis.clickFocus <- function() {
+trellis.clickFocus <- function(...) {
 	layoutMatrix <- trellis.currentLayout()
 	currVpp <- current.vpPath()
 	if (!is.null(currVpp)) { upViewport(currVpp$n) }
@@ -806,23 +890,11 @@ trellis.clickFocus <- function() {
 		focusCol <- focusRow <- 0
 		trellis.unfocus()
 	} else {
-		trellis.focus("panel", focusCol, focusRow)
+		trellis.focus("panel", focusCol, focusRow, ...)
 	}
 	invisible(list(col=focusCol, row=focusRow))
 }
 
-# this function by Deepayan Sarkar
-packet.panel.page <- function(n)
-{
-   ## returns a function that when used as the 'packet.panel'
-   ## argument in print.trellis plots page number 'n' only
-   function(layout, page, ...) {
-       stopifnot(layout[3] == 1)
-       packet.panel.default(layout = layout,
-                            page = page + n - 1,
-                            ...)
-   }
-}
 
 latticeUnLog <- function(xlim, ylim, scalesArg) {
 	# x scale
@@ -853,6 +925,7 @@ latticeUnLog <- function(xlim, ylim, scalesArg) {
 }
 
 latticeLogBase <- function(x) {
+	x <- eval(x)
 	if (is.null(x) || identical(x, FALSE)) {
 		return(NULL)
 	}
@@ -866,12 +939,14 @@ as.POSIXct.numeric <- function(x) {
 	structure(as.numeric(x), class = c("POSIXt", "POSIXct"))
 }
 
-evalCallArgs <- function(myCall, envir=parent.frame(), pattern=".") {
+evalCallArgs <- function(myCall, envir=parent.frame(), pattern=T) {
 	for (i in seq(along=myCall)) {
 		if ((mode(myCall) %in% "call") && (i == 1)) {
 			next # don't eval function itself
 		}
-		if (length(grep(pattern, deparse(myCall[[i]])))>0) {
+		if (isTRUE(pattern)) {
+			myCall[[i]] <- eval(myCall[[i]], envir)
+		} else if (any(grep(pattern, deparse(myCall[[i]])))) {
 			myCall[[i]] <- eval(myCall[[i]], envir)
 		} else if (mode(myCall[[i]]) %in% c("call","list")) {
 			myCall[[i]] <- evalCallArgs(myCall[[i]], envir=envir,
@@ -881,3 +956,81 @@ evalCallArgs <- function(myCall, envir=parent.frame(), pattern=".") {
 	return(myCall)
 }
 
+## The following functions by Deepayan Sarkar Deepayan.Sarkar@R-project.org
+
+packet.panel.page <- function(n)
+{
+   ## returns a function that when used as the 'packet.panel'
+   ## argument in print.trellis plots page number 'n' only
+   function(layout, page, ...) {
+       stopifnot(layout[3] == 1)
+       packet.panel.default(layout = layout,
+                            page = page + n - 1,
+                            ...)
+   }
+}
+
+highlightOne <-
+   function(pargs = trellis.panelArgs(),
+            threshold = 18,
+            col = 1, pch = 16, ...)
+{
+   #cat("click to choose one point to highlight", fill = TRUE)
+
+   ll <- grid.locator(unit = "npc")
+   if (is.null(ll)) return(FALSE)
+
+   nvars <- length(pargs$z)
+
+   ## which subpanel
+   colpos <- ceiling(convertUnit(ll$x, "npc", valueOnly = TRUE) * nvars)
+   rowpos <- ceiling(convertUnit(ll$y, "npc", valueOnly = TRUE) * nvars)
+   if (rowpos == colpos) return(TRUE)
+   subpanel.name <- paste("subpanel", colpos, rowpos, sep = ".")
+
+   ## coordinates of click in subpanel
+   ll$x <- nvars * (ll$x - unit((colpos-1) / nvars, "npc"))
+   ll$y <- nvars * (ll$y - unit((rowpos-1) / nvars, "npc"))
+
+   ## get to that viewport, so we can convert units
+   depth <- downViewport(subpanel.name)
+   xnative <- convertX(ll$x, "native", TRUE)
+   ynative <- convertY(ll$y, "native", TRUE)
+
+   ## find nearest point in data (replicate steps in panel.identify)
+
+   xpoints <- convertX(unit(xnative, "native"), "points", TRUE)
+   ypoints <- convertY(unit(ynative, "native"), "points", TRUE)
+
+   data.xp <- convertX(unit(pargs$z[, colpos], "native"), "points", TRUE)
+   data.yp <- convertY(unit(pargs$z[, rowpos], "native"), "points", TRUE)
+
+   pdists <- sqrt((data.xp - xpoints)^2 + (data.yp - ypoints)^2)
+
+   if (min(pdists, na.rm = TRUE) > threshold)
+   {
+       warning("no points within ", threshold, " points of click")
+       upViewport(depth)
+   }
+  else
+   {
+       w <- which.min(pdists)
+       ## print(pargs$z[w,])
+       upViewport(depth)
+       for (row in 1:nvars)
+       for (column in 1:nvars)
+           if (row != column)
+           {
+               subpanel.name <-
+                   paste("subpanel",
+                         column, row, sep = ".")
+               depth <- downViewport(subpanel.name)
+               panel.points(x = pargs$z[w, column],
+                            y = pargs$z[w, row],
+                            pch = pch, col = col,
+                            ...)
+               upViewport(depth)
+           }
+   }
+   return(TRUE)
+}
